%% Nick Sullivan 2017, The University of Adelaide


classdef GAAlgs
    methods(Static)
            
            
            % Super useful guide:
    %(Sastry, Kumara, David E. Goldberg, and Graham Kendall. "Genetic
    %algorithms." Search methodologies. Springer US, 2014. 93-117.)
    
    % POSSIBLE OPERATORS
    %(Genetic Algorithm for the Traveling Salesman Problem using Sequential Constructive Crossover Operator )
    %(Genetic Algorithms for the Travelling Salesman Problem: A Review of Representations and Operators)
    %(GELS-GA: Hybrid Metaheuristic Algorithm for Solving Multiple Travelling Salesman Problem)
    % Single-Point exchange operator
    
        % Performs crossover function from this paper:
        % "A new crossover approach for solving the multiple travelling
        % salesmen problem using genetic algorithms"
        function crossover(mother, father, child_ptr)
            child_gene = mother.gene;
            num_r  = mother.num_robots;
            num_ta = mother.num_tasks;
            % 1) Select a random sub tour for each robot from Mum. Add all
            % unselected nodes into a vector.
            tasks = mother.task_paths;
            tools = mother.tool_paths;
            sub_tasks = cell(num_r, 1);
            sub_tools = cell(num_r, 1);
            unselected_tasks = [];
            unselected_tools = [];
            for r = 1:num_r
                % Select indices.
                len = length(tasks{r});
                if len < 1
                    continue;
                end
                i1 = randi([1 len]);
                i2 = randi([i1 len]);
                % Copy selected into the sub tour, and add unselected to a
                % vector.
                sub_tasks{r} = tasks{r}(i1:i2);
                sub_tools{r} = tools{r}(i1:i2);
                un_tasks = tasks{r};
                un_tools = tools{r};
                un_tasks(i1:i2) = [];
                un_tools(i1:i2) = [];
                unselected_tasks = [unselected_tasks un_tasks];
                unselected_tools = [unselected_tools un_tools];
            end
            % 2) Take all unselected genes, and match the order in Dad.
%             [unselected_tasks, ind1, ind2] = intersect(father.gene(1:num_ta), unselected_tasks, 'stable');
% %             unselected_tools = unselected_tools(ind2);   % use this to keep mums task-tools
%             unselected_tools = father.gene(num_ta + ind1); % use this to use dads task-tools
            [~, locb] = ismember(unselected_tasks, father.gene(1:num_ta));             % this is faster
            [~, ind] = sort(locb);
            unselected_tasks = unselected_tasks(ind);
            unselected_tools = father.gene(num_ta + locb);
            unselected_tools = unselected_tools(ind);

            % 3) Split the unselected genes randomly between robots.
            num_unalloc = length(unselected_tasks);   % how many to allocate
            allocs = zeros(num_r, 1);          % unselected task allocations
            r_order = randperm(num_r);         % being allocated first increases likelihood of getting all
            for r = r_order(1:end-1)
                if num_unalloc > 0
                    alloc = randi([1 num_unalloc]);
                    allocs(r) = alloc;
                    num_unalloc = num_unalloc - alloc;
                end
            end
            allocs( r_order(end) ) = num_unalloc;  % last robot gets the rest
            % 4) Fill in the child using selected and unselected.
            child_i = 1;
            unselected_i = 1;
            for r = 1:num_r
               % Start with selected.
               tasks_to_add = sub_tasks{r};
               tools_to_add = sub_tools{r};
               for i = 1:length(tasks_to_add)
                   child_gene(child_i) = tasks_to_add(i);
                   child_gene(child_i+num_ta) = tools_to_add(i);
                   child_i = child_i + 1;
               end
               % Then add unselected.
               len = allocs(r);
               tasks_to_add = unselected_tasks(unselected_i:unselected_i+len-1);
               tools_to_add = unselected_tools(unselected_i:unselected_i+len-1);
               unselected_i = unselected_i + len;
               for i = 1:length(tasks_to_add)
                   child_gene(child_i) = tasks_to_add(i);
                   child_gene(child_i+num_ta) = tools_to_add(i);
                   child_i = child_i + 1;
               end
            end

            % 5) Set the path lengths (right side of chromosome).
            for r = 1:num_r
               len = allocs(r) + length(sub_tasks{r});
               child_gene(num_ta*2 + r) = len;
            end
            
            child_ptr.setGene(child_gene);
        end

        % Mutation - internal swap. Two tasks are swapped within a robots
        % path.
        function mutateInternalSwap( ga_solution, new_ga_solution )
            num_r = ga_solution.num_robots;
            % Pick a robot.
            r = floor( rand*num_r + 1 );
            tasks = ga_solution.task_paths;
            tools = ga_solution.tool_paths;
            path_len = length(tasks{r});
            if path_len < 2
                return;
            end
            % Flip tasks and tools.
            i1 = floor( rand*path_len + 1 );
            i2 = floor( rand*path_len + 1 );
            tasks{r}([i1 i2]) = tasks{r}([i2 i1]);
            tools{r}([i1 i2]) = tools{r}([i2 i1]);
            % Set it.
            new_ga_solution.setPaths(tasks, tools);
        end
        
        
        % Mutation - external swap. Two tasks are swapped between robots.
        function mutateExternalSwap( ga_solution, new_ga_solution )
            num_r = ga_solution.num_robots;
            
            % Pick the robots.
            r1 = floor( rand*num_r + 1 );
            r2 = floor( rand*num_r + 1 );
            tasks = ga_solution.task_paths;
            tools = ga_solution.tool_paths;
            path_len1 = length(tasks{r1});
            path_len2 = length(tasks{r2});
            % Check they can be swapped.
            if path_len1 < 1 || path_len2 < 1 || r1 == r2
                return;
            end
            
            % Swap two tasks..
            i1 = floor( rand*path_len1 + 1 );
            i2 = floor( rand*path_len2 + 1 );
            
            temp = tasks{r1}(i1);
            tasks{r1}(i1) = tasks{r2}(i2);
            tasks{r2}(i2) = temp;
            
            temp = tools{r1}(i1);
            tools{r1}(i1) = tools{r2}(i2);
            tools{r2}(i2) = temp;
            
            % Set.
            new_ga_solution.setPaths(tasks, tools);
        end
        
        
        % Mutation - reverse. Tasks in a random section are reversed.
        % objective - [1|2] 1=minisum, 2=minimax
        function mutateReverse( ga_solution, new_ga_solution )
            num_r = ga_solution.num_robots;
            
            r = randi([1 num_r]);
            tasks = ga_solution.task_paths;
            tools = ga_solution.tool_paths;
            path_len = length(tasks{r});
            if path_len < 2
                return;
            end
            
            % Flip tasks and tools.
            ind = sort(randi([1 path_len], 1, 2));
            i1 = ind(1);
            i2 = ind(2);
            tasks{r}(i1:i2) = tasks{r}(i2:-1:i1);
            tools{r}(i1:i2) = tools{r}(i2:-1:i1);
            
            % Update.
            new_ga_solution.setPaths(tasks, tools);
        end
        
        % Mutation - transfer. A task is given from one robot to another.
        function mutateTransfer( ga_solution, new_ga_solution )
            num_r = ga_solution.num_robots;
            
            r1 = randi([1 num_r]);
            r2 = randi([1 num_r]);
            if r1 == r2
                return;
            end
            tasks = ga_solution.task_paths;
            tools = ga_solution.tool_paths;
            path1_len = length(tasks{r1});
            path2_len = length(tasks{r2});
            if path1_len < 1
                return;
            end
            
            ind1 = randi([1 path1_len]);
            ind2 = randi([1 path2_len+1]);
            
            ta1 = tasks{r1}(ind1);
            to1 = tools{r1}(ind1);
            
            tasks{r2} = RobotAlgs.insertValueIntoVector( tasks{r2}, ta1, ind2 );
            tools{r2} = RobotAlgs.insertValueIntoVector( tools{r2}, to1, ind2 );
            tasks{r1}(ind1) = [];
            tools{r1}(ind1) = [];
            
            new_ga_solution.setPaths(tasks, tools); 
        end
        
        
        % Mutation - tool change.
        function mutateToolChange( ga_solution, new_ga_solution )
            num_ta = ga_solution.num_tasks;
            num_to = ga_solution.num_tools;
            if num_to < 2
                return;
            end
            t = randi(num_ta);
            type = ga_solution.task_list(t,2);
            options = [ (1:num_to)' ga_solution.tool_expertise(:,type)];
            options = options( options(:,2) > 0, : );
            tool = options( randi(size(options,1)), 1 );
            
            new_gene = ga_solution.gene;
            new_gene(num_ta + t) = tool;
            
            new_ga_solution.setGene(new_gene);
        end
        
        
         % Mutation - batch tool change. Change a random path section to use 
        % the same tools. 
        function mutateMatchTool( ga_solution, new_ga_solution )
            num_r = ga_solution.num_robots;
            num_ta = ga_solution.num_tasks;
            
            % Pick a robot.
            r = randi( num_r );
            tasks = ga_solution.task_paths;
            tools = ga_solution.tool_paths;
            path_len = length(tasks{r});
            if path_len < 1
                return;
            end
            
            % Pick a subsection.
            ind = sort(randi([1 path_len], 1, 2));
            i1 = ind(1);
            i2 = ind(2);
            % Pick a task within that subsection.
            ind = randi([i1 i2]);
            tool = tools{r}(ind);
            % Set all tools in the subsection to match the chosen one.
            tools{r}(i1:i2) = tool;
            
            new_ga_solution.setPaths(tasks, tools);
        end
        
        % Improves the gene using 2-opt. Additional options for a single
        % operation (faster but worse), and if we should only improve the
        % longest path (minimax isn't improved if we improve shorter
        % paths).
        function ga_solution = perform2Opt( ga_solution, one_shot_improvement, only_longest_robot )
            % Prep.
            new_ga_solution = ga_solution.copy();
            num_r = ga_solution.num_robots;
            % Get the robot with the longest path.
            [~, r_ind] = max(ga_solution.costs);
            % Perform 2-opt improvement for each robot.
            for r = 1:num_r
                if only_longest_robot
                   if r ~= r_ind
                       continue;
                   end
                end
                improved = true;
                while improved
                    improved = false;
                    % Pick a start node.
                    len = length(ga_solution.task_paths{r});
                    for i=1:len-1
                        % Pick an end node.
                        for k=i+1:len
                            % Swap.
                            new_tasks = ga_solution.task_paths;
                            new_tools = ga_solution.tool_paths;
                            new_tasks{r} = GAAlgs.twoOptSwap(new_tasks{r}, i, k);
                            new_tools{r} = GAAlgs.twoOptSwap(new_tools{r}, i, k);
                            new_ga_solution.setPaths(new_tasks, new_tools);
                            % Keep if it's better.
                            if (new_ga_solution.minimax < ga_solution.minimax) || ...
                               (new_ga_solution.minimax == ga_solution.minimax && new_ga_solution.minisum < ga_solution.minisum)
                                ga_solution.setPaths(new_tasks, new_tools);
                                improved = true;
                            end
                        end
                    end
                    if one_shot_improvement
                       break; 
                    end
                end
            end
        end
        
        
        % From wikipedia. p1 must be < p2
        function new_path = twoOptSwap(path, p1, p2)
            new_path = path;
            new_path(p1:p2) = new_path(p2:-1:p1);
        end
    end
end



