%% Nick Sullivan 2017, The University of Adelaide


classdef GASolution < handle  % Inheriting from 'handle' makes this return a pointer.
    properties
        
        % Stored once at beginning.
        num_tasks    
        num_tools
        num_robots
        num_tool_racks
        
        robot_list           % [id x y
                             %  id x y]      start positions of each robot
        task_list            % [id type x y base_cost
                             %  id type x y base_cost]  types, positions, and costs to complete for each task.
        tool_rack_list       % [id x y cost
                             %  id x y cost] locations and costs of tool racks.
        robot_expertise      % [expforT1 expforT2 expforT3
                             %  expforT1 expforT2 expforT3] robot's base (no tools) expertise for each task type
        tool_expertise       % [expforT1 expforT2 expforT3
                             %  expforT1 expforT2 expforT3] tool's bonus expertise for each task type
        travel_costs         % [c11 c12 c13 c14 ...  the cost matrix to travel from one node to another
                             %  c21 c22 c23 c24 ...]  The first R entries are
                             %                 robot start positions, the next Ta entries are
                             %                 the tasks, the end To entries are tool racks.
        travel_costs_swap    % The same as travel_costs, but includes the distance from tool-swapping between
                             % the two nodes. Does not include tool rack entries.   
        travel_racks_swap    % The rack IDs used in travel_costs_swap.
                             
        return_home          % Whether or not the robots must return to their starting point.
        
        % Do not alter these values directly!
        gene    % In the form [ Tasks | Tools | Robots ] e.g.
                % [ 1 2 3 4 | 1 1 2 2 | 2 2 ] means that robots 1 and 2
                % both have path lengths of 2 (section at end). This
                % indicates that the tasks and tools are segmented in half,
                % so Robot 1 will complete task1 with tool1, then will
                % complete task2 with tool1. Robot 2 will complete task3
                % with tool2, then will complete task4 with tool2.
                % Another example:
                % [1 3 2 4 | 1 2 1 1 | 3 1 ] means that robot 1 completes 3
                % tasks, and robot 2 completes one task (section at end).
                % robot 1 will complete task1 with tool1, then task3 with
                % tool1, then task2 with tool2. Robot 2 will complete task4
                % with tool1.
    
        task_paths      % { [t1 t2 t3] } One cell for each robot. e.g.
                        % task_paths{1} returns the path for robot 1.
        tool_paths      % { [t1 t2 t3] } One cell for each robot, returns the tools being used for each path.
        
        % Do not alter these values directly!
        costs           % [r1 r2 r3], the path costs for each robot
        
        minimax         % the cost according to minimax
        minisum         % the cost according to minisum

    end
    
    methods(Access=private)
        
        % Populates the 'travel_costs' property using euclidean distance.
        % travel_costs - [c11 c12 c13 ...  The travel cost between two
        %                 c21 c22 c23 ...] nodes. The first R entries are
        %                 robot start positions, the next Ta entries are
        %                 the tasks, the end To entries are tool racks.
        function populateTravelCosts(this)
            num_entries = this.num_robots+this.num_tasks;
            
            x = [this.robot_list(:,2); this.task_list(:,3)];
            y = [this.robot_list(:,3); this.task_list(:,4)];
            if ~isempty(this.tool_rack_list)
                num_entries = num_entries + this.num_tool_racks;
                x = [x; this.tool_rack_list(:,2)];
                y = [y; this.tool_rack_list(:,3)];
            end
            
            this.travel_costs = zeros(num_entries);
            this.travel_costs_swap = zeros(num_entries);
            this.travel_racks_swap = zeros(num_entries);
            for i=1:num_entries
               for j=1:num_entries
                   % From i to j
                   dx = x(j) - x(i);
                   dy = y(j) - y(i);
                   dist = sqrt( dx^2 + dy^2 );
                   this.travel_costs(i, j) = dist;
                   % Calculate best rack for i -> rack -> j
                   if this.num_tool_racks > 0
                       dx = this.tool_rack_list(:,2) - x(i);
                       dy = this.tool_rack_list(:,3) - y(i);
                       dx2 = x(j) - this.tool_rack_list(:,2);
                       dy2 = y(j) - this.tool_rack_list(:,3);
                       pos_to_rack = sqrt( dx.^2 + dy.^2 );
                       swap = this.tool_rack_list(:,4);
                       rack_to_pos = sqrt( dx2.^2 + dy2.^2 );
                       [dist, rack_id] = min( pos_to_rack + swap + rack_to_pos );
                       this.travel_costs_swap(i,j) = dist;
                       this.travel_racks_swap(i,j) = rack_id;
                   end
               end
            end
        end
        
        % If the easy-access values have been changed, use this to
        % recalculate the gene vector.
        function updateGene(this)
            num_r = this.num_robots;
            num_ta = this.num_tasks;
            new_gene = zeros(1,num_ta*2 + num_r);
            index = 1;
            for r = 1:num_r
                tasks = this.task_paths{r};
                tools = this.tool_paths{r};
                new_gene(index:index+length(tasks)-1) = tasks;   % tasks
                new_gene( num_ta+index:num_ta+index+length(tasks)-1 ) = tools; % tools
                new_gene(num_ta*2 + r) = length(tasks);    % lengths
                index = index + length(tasks);
            end
            this.gene = new_gene;
        end
        
        % If the gene has been changed, use this to update the easy-access
        % properties.
        function updateEasyAccess(this)
            num_r = this.num_robots;
            num_ta = this.num_tasks;
            
            tasks = cell(num_r, 1);
            tools = cell(num_r, 1);
            last_index = 0;
            for r = 1:num_r
                len = this.gene(num_ta*2 + r);
                tasks{r} = this.gene(last_index+1:last_index+len);
                tools{r} = this.gene(num_ta+last_index+1:num_ta+last_index+len);
                last_index = last_index+len;
            end
            
            this.task_paths = tasks;
            this.tool_paths = tools;
        end
        
        % Populates the 'cost' property of this gene. Uses easy access
        % variables so make sure they're up to date.
        function calculateGeneCost(this)
            % Split into robot paths.
            num_r = this.num_robots;
            this.costs = zeros(num_r, 1);
            this.minisum = 0;
            this.minimax = 0;
            for r = 1:num_r
                r_cost = 0;
                tasks = this.task_paths{r};
                tools = this.tool_paths{r};
                % Calculate cost.
                equipped_tool = 1;                % 'no' tool
                pos_ind = r;
                for t = 1:length(tasks)
                    tool = tools(t);
                    task = tasks(t);
                    type = this.task_list(task,2);
                    if tool ~= equipped_tool
                        equipped_tool = tool;
                        travel_cost = this.travel_costs_swap(pos_ind, num_r + task);
                    else
                        travel_cost = this.travel_costs(pos_ind, num_r + task);
                    end
                    expertise = this.robot_expertise(r, type) + this.tool_expertise(tool,type);
                    completion_cost = this.task_list(task,5) / expertise;
                    r_cost = r_cost + travel_cost + completion_cost;
                    pos_ind = num_r+task;
                end
                % Add returning home.
                if this.return_home
                    if equipped_tool ~= 1
                        travel_cost = this.travel_costs_swap(pos_ind, r);
                    else
                        travel_cost = this.travel_costs(pos_ind, r);
                    end
                    r_cost = r_cost + travel_cost;
                end
                % Update cost.
                this.costs(r) = r_cost;
                this.minisum = this.minisum + r_cost;
                if this.minimax < r_cost
                    this.minimax = r_cost;
                end
            end
        end
    end
    
    methods
        % Constructor. Randomly fill the gene until it is a solution.
        % tool_seed = [1|2] 1:pick a random valid tool, 2:pick the best tool
        function obj = GASolution(robot_list, task_list, tool_rack_list, robot_expertise, tool_expertise, return_home, tool_seed, auction_seed, objective, transform)
            % If we provide no arguments, do nothing.
            if nargin == 0
               return; 
            end
            % Populate properties.
            num_robots = size(robot_list, 1);
            num_tasks = size(task_list, 1);
            num_tools = size(tool_expertise, 1);
            num_tool_racks = size(tool_rack_list, 1);
            obj.num_robots = num_robots;
            obj.num_tasks = num_tasks;
            obj.num_tools = num_tools;
            obj.num_tool_racks = num_tool_racks;
            obj.robot_list = robot_list;
            obj.task_list = task_list;
            obj.tool_rack_list = tool_rack_list;
            obj.robot_expertise = robot_expertise;
            obj.tool_expertise = tool_expertise;
            obj.return_home = return_home;
            
            if ~auction_seed
                % Randomly fill in tasks from 1 to num_tasks.
                g_tasks = randperm(num_tasks);
                % Randomly fill in tools with valid options from 1 to num_tools.
                g_tools = floor( rand(1, num_tasks) * num_tools + 1 );
                for i=1:num_tasks
                    task = g_tasks(i);
                    type = task_list(task,2);
                    if tool_seed == 1
                        options = [ (1:num_tools)' tool_expertise(:,type)];
                        options = options( options(:,2) > 0, : );
                        tool = options( randi(size(options,1)), 1 );
                    end
                    if tool_seed == 2
                        [~,tool] = max(tool_expertise(:,type)); % ...or use the best tool as a seed.
                    end
                    g_tools(i) = tool;
                end
                % Randomly distribute the tasks to the robots.
                g_robots = zeros(1, num_robots);
                tasks_remain = num_tasks;
                for r = 1:num_robots-1
                    allocate = floor( rand() * (tasks_remain+1) );
                    g_robots(r) = allocate;
                    tasks_remain = tasks_remain - allocate;
                end
                g_robots(num_robots) = tasks_remain;
                % Put them together.
                obj.gene = [g_tasks g_tools g_robots];

                % Update other properties.
                obj.updateEasyAccess();
                obj.populateTravelCosts();
                obj.calculateGeneCost();
            end
            % Experimental, setting seed from auction.
            if auction_seed
%                 objective = 1;  % manually alter
%                 transform = true; % manually alter
                
                s = State();
                s.return_home = return_home;
                s.robot_list = robot_list;
                s.task_list = task_list;
                s.tool_rack_list = tool_rack_list;
                s.robot_expertise = robot_expertise;
                s.tool_expertise = tool_expertise;
                
                preallocate = 500;
                s.state_index = 1;
                s.allocated_tasks    = cell(preallocate, 1);
                s.allocated_tasks{1} = [];
                s.unallocated_tasks    = cell(preallocate, 1);
                s.unallocated_tasks{1} = s.task_list(:,1);
                s.path_plan    = cell(preallocate, 1);
                s.path_plan{1} = cell( size(s.robot_list, 1), 1 );   % empty path plans
                
                s.populateTravelCosts();
                
                if transform
                    s2 = s.copy();
                    [s2, E_lim, E_max] = State.transformToMTSP(s2);
                    s2 = State.solveStates(s2, objective, 1, false);
                    s2 = State.transformFromMTSP(s2);
                    s.path_plan = s2.path_plan;
                    s.minimax_cost = s2.minimax_cost;
                    s.minisum_cost = s2.minisum_cost;
                    s.state_index = s2.state_index;
                else 
                    s = State.solveStates(s, objective, 1, false);
                end

                obj = s.convertToGASolution();
            end
            % End experiment.
            
        end
        
        % Changes the 'gene' value.
        function setGene(this, gene)
            this.gene = gene;
            this.updateEasyAccess();
            this.calculateGeneCost();
        end
        
        % Changes the task paths and tool paths.
        function setPaths(this, task_paths, tool_paths)
           this.task_paths = task_paths;
           this.tool_paths = tool_paths;
           this.updateGene();
           this.calculateGeneCost();
        end
        
        % Make a copy of a handle object. Use sparingly.
        function new = copy(this)
            % Instantiate new object of the same class.
            new = GASolution();
            % Copy all non-hidden properties.
            p = properties(this);
            for i = 1:length(p)
                new.(p{i}) = this.(p{i});
            end
        end
        
        % Plots the state.
        % objective -[1|2] 1 = minisum, 2 = minimax
        function plotGene(this, objective)
            thick_mode = true;  % change this if you want
            if thick_mode
                task_size = 40;
                tool_size = 120;
                robot_size = 140;
                line_width = 2;
            else
                task_size = 25;
                tool_size = 50;
                robot_size = 40;
                line_width = 1;
            end
            
            hold on;
            num_r = this.num_robots;
            colours = lines(num_r+1);
                
            % Get the tasks.
            tasks = this.task_paths;
            tools = this.tool_paths;
            
            % Plot robots.
            for r = 1:num_r
                len = length(tasks{r});
                x = zeros(len+1, 1);
%                 if this.return_home
%                     x = zeros(len+2, 1);
%                 end
                y = x;
                x(1) = this.robot_list(r,2);
                y(1) = this.robot_list(r,3);
                equipped_tool = 1;
                ind = 2;
                pos_ind = r;
                for i = 1:len
                    tool = tools{r}(i);
                    task = tasks{r}(i);
                    if tool ~= equipped_tool
                        rack = this.travel_racks_swap(pos_ind,num_r+task);
                        x(ind) = this.tool_rack_list(rack,2);
                        y(ind) = this.tool_rack_list(rack,3);
                        equipped_tool = tool;
                        ind = ind + 1;
                    end
                    x(ind) = this.task_list(task, 3);
                    y(ind) = this.task_list(task, 4);
                    ind = ind + 1;
                    pos_ind = num_r + task;
                end
                if this.return_home
                    if equipped_tool ~= 1
                        rack = this.travel_racks_swap(pos_ind,r);
                        x(ind) = this.tool_rack_list(rack,2);
                        y(ind) = this.tool_rack_list(rack,3);
                        equipped_tool = 1;
                        ind = ind + 1;
                    end
                    x = [x; this.robot_list(r,2)];
                    y = [y; this.robot_list(r,3)];
                end
                plot(x, y, 'Color', colours(r,:), 'Linewidth', line_width);
                scatter(x(1), y(1), robot_size, 'MarkerFaceColor', colours(r,:), 'MarkerEdgeColor', [0 0 0], 'Linewidth', line_width);
            end
             
            % Plot tasks.
            tasks_by_colour = true;
            if tasks_by_colour
                for i = 1:max(this.task_list(:,2))
                    x = this.task_list(:, 3);
                    y = this.task_list(:, 4);
                    x = x( this.task_list(:,2) == i );
                    y = y( this.task_list(:,2) == i );
    %                 scatter(x, y, task_size, 'filled', 'MarkerEdgeColor',[0 0 0], 'MarkerFaceColor',[0 0 0]);
                    scatter(x, y, task_size, 'filled');
                end
            else
                scatter_types = ['*';'+';'d'];
                for i = 1:max(this.task_list(:,2))
                    x = this.task_list(:, 3);
                    y = this.task_list(:, 4);
                    x = x( this.task_list(:,2) == i );
                    y = y( this.task_list(:,2) == i );
                    scatter(x, y, task_size, scatter_types(i), 'MarkerEdgeColor',[0 0 0], 'MarkerFaceColor',[0 0 0],'Linewidth', 1);
                end
            end
            
            % Plot the tool change.
            if ~isempty(this.tool_rack_list)
                x = this.tool_rack_list(:,2);
                y = this.tool_rack_list(:,3);
%                 scatter(x, y, 50, 'square', 'MarkerFaceColor', [0 0 0], 'MarkerEdgeColor', [0 0 0]);
%                 scatter(x, y, task_size, scatter_types(i), 'MarkerEdgeColor',[0 0 0], 'MarkerFaceColor',[0 0 0],'Linewidth', 1);
                scatter(x, y, tool_size, 'square', 'MarkerFaceColor', [0 0 0], 'MarkerEdgeColor', [0 0 0]);
            end


            
%             xlim([-10 210]);
%             ylim([-60 40]);
            if objective == 2
                title_str = ['GA MiniMax: ' num2str(this.minimax)];
            end
            if objective == 1
                title_str = ['GA MiniSum: ' num2str(this.minisum)];
            end
            title(title_str);
            
        end
    end
end



