%% Nick Sullivan 2017, The University of Adelaide


classdef RobotAlgs
    methods(Static)
        
        % Returns the state ID with the given code.
        function state_id = findState(code, states)
            state_id = 0;
            for i = 1:length(states)
                s = states(i);
                c = s.code;
                if length(code) ~= length(c)  % must be same length
                    continue;
                end
                if all( code == c )  % all entries must match
                    state_id = i;
                    return;
                end
            end
        end
        
        
        function [tool_types] = getToolsWithBestEvennessMetric( robot_id, robot_expertise, tool_expertise, equipped_tools, unallocated_tasks, task_list )
            % Pick the tool that will make the evenness metric the smallest.
            num_tools = size(tool_expertise, 1);
            em = RobotAlgs.calculateEvennessMetric( robot_id, robot_expertise, tool_expertise, equipped_tools, unallocated_tasks, task_list  );
            % Add tool numbers to keep track.
            em = [(1:num_tools)' em];
            % Remove all those with greater than minimum infs.
            min_infs = min(em(:,2));
            em( em(:,2) > min_infs, : ) = [];
            % Select the ones with the minimum spread.
            m = min(em(:,3));
            tool_indices = em(:,3) <= m;
            tool_types = em(tool_indices, 1);
        end
        
        % Chooses the desired tool for a given robot.
        % id - [int]
        % equipped_tools - [toolid toolid ...] tools equipped by each robot.
        % state - obj
        function [tool_type] = selectBestTool( id, equipped_tools, state )
            equipped_tool = equipped_tools(id);
            % If the last thing in a path is a tool swap, disallow any bids on
            % tools for that robot. (MASSIVELY reduces number of states)
            path = state.path_plan{state.state_index}{id};
            if ~isempty(path)
                if path(end) < 1
                    tool_type = equipped_tool;
                    return;
                end
            end
            % If we have no other option but to swap, then we must swap.
%             num_tools = size(state.tool_expertise, 1);
%             predictors = state.calculateMLPredictors(id, state.state_index);
%             tasks_left = predictors(3);
%             must_swap = false;
%             if tasks_left <= 0
%                 must_swap = true;
%             end      
            
            % Pick the tool that will make the evenness metric the
            % smallest.
            tools = RobotAlgs.getToolsWithBestEvennessMetric( id, state.robot_expertise, state.tool_expertise, equipped_tools, state.unallocated_tasks{state.state_index}, state.task_list  );
            tool_type = tools(1);
            
        end
        
        % Calculates how even robot expertise is allocated amongst remaining tasks.
        % Returns an array for the evenness metric if the provided robot were to
        % change to each tool.
        % id  - [int]
        % robot_expertise - [robot1ExpForT1 robot1ExpForT2 robot1ExpForT3
        %                    robot2ExpForT1 robot2ExpForT2 robot2ExpForT3 ]
        % tool_expertise - [tool1ExpForT1 tool1ExpForT2 tool1ExpForT3
        %                   tool2ExpForT1 tool2ExpForT2 tool2ExpForT3 ]
        % equipped_tools - [id1 id2 ...]
        % unallocated_tasks - [id1 id2 ...]
        % task_list - [id type x y base_cost
        %              id type x y base_cost]
        function [em] = calculateEvennessMetric( id, robot_expertise, tool_expertise, equipped_tools, unallocated_tasks, task_list )
            num_task_types = size(robot_expertise, 2);
            num_robots = size(robot_expertise, 1);
            num_tools = size(tool_expertise, 1);
            % Count how many tasks there are of each type.
            task_counts = zeros(1, num_task_types);
            for t = 1:length(unallocated_tasks)
                t_id = unallocated_tasks(t);
                type = task_list(t_id, 2);
                task_counts(type) = task_counts(type) + 1;
            end
            
            % Calculate the total expertise of all base robot expertise + all OTHER
            % robots equipping their tools.
            total_expertise = zeros( 1, num_task_types);
            for r = 1:num_robots
                total_expertise = total_expertise + robot_expertise(r,:);
                if r ~= id
                    tool = equipped_tools(r);
                    total_expertise = total_expertise + tool_expertise(tool,:);
                end
            end
            % Add each tool to the main robot, and calculate the evenness metric.
            % The metric uses the following formula:
            % Count number of tasks of each type.
            % Sum expertise of all robots for each task type.
            % Divide the number of tasks by the expertise for each type.
            % Calculate the standard deviation of those ratios.
            
            % The metric is a tuple of 2 components, the first the number of
            % infinities that occur when dividing (i.e. zero expertise for nonzero
            % tasks). The second is the standard deviation of the non-infinites.
            
            em = zeros(num_tools, 2);
            for t = 1:num_tools
                expertise = total_expertise + tool_expertise(t,:);
                ratios = task_counts ./ expertise;
                % Remove 0 divided by 0.
                ratios(isnan(ratios)) = [];
                % Count infs.
                num_infs = sum( isinf(ratios) );
                ratios(isinf(ratios)) = [];
                % Calculate the spread.
%                 spread = std(ratios);
                spread = sum(ratios);       % TEST
                if isempty(ratios)
                    spread = 0;
                end
                if isnan(spread)
                    spread = inf;
                end
                
                em(t,:) = [num_infs spread];
            end
        end
        
        % task_bids - [bid1 bid2 ...
        %              bid1 bid2 ...] row is robot id, col is task_id
        % tool_bids - [bid1 bid2 ...
        %              bid1 bid2 ...] row is robot_id, col is tool_id
        % r_id - [int] id of winning robot
        % is_task - [bool] true if they've won a task, false if they've won a tool
        % t_index - [int] index of task or tool won.
        % tool_bid_option - [1|2]  1:  tool bids are the cost of tool swap + completing one task
        %                          2:  tool bids are binary and allocated first - it is up to the robot to want a tool
        function [r_id, is_task, t_index] = auctionDecide(task_bids, tool_bids, tool_bid_option)
            num_robots = size(task_bids,1);
            num_tasks = size(task_bids,2);
            num_tools = size(tool_bids,2);
            
            resolution_type = 1;        % 1: lowest bid wins
                                        % 2: least contested bid wins
            
            % 1) Overall minimum:
            if resolution_type == 1
                % Find the overall minimum task bid.
                [winning_task_bid, task_id] = min(min(task_bids,[],1));
                [~,winning_task_robot] = min(task_bids(:,task_id));
%                 if isempty(min_task_bid)
%                     min_task_bid = inf;
%                 end
                % Find the overall minimum tool bid.
                [winning_tool_bid, tool_id] = min(min(tool_bids,[],1));
                [~,winning_tool_robot] = min(tool_bids(:,tool_id));
%                 if isempty(min_tool_bid)
%                     min_tool_bid = inf;
%                 end
            end
            
            % 2) Least contested bid:
            if resolution_type == 2
                winning_task_bid = -1;
                winning_task_dbid = 0;
                winning_task_robot = 0;
                task_id = 0;
                for t = 1:num_tasks
                    bids = task_bids(:,t);
                    [min_tasks_1, id] = min(bids);  % get lowest
                    bids(id) = [];
                    min_tasks_2 = min(bids);  % get second lowest
                    dbid = min_tasks_2 - min_tasks_1;
                    if isnan(dbid)
                        dbid = 0;
                    end
                    if dbid > winning_task_dbid
                        winning_task_bid = min_tasks_1;
                        winning_task_robot = id;
                        winning_task_dbid = dbid;
                        task_id = t;
                    end
                end
                if winning_task_bid < 0
                    winning_task_bid = inf;
                end
                winning_tool_bid = -1;
                winning_tool_dbid = -1;
                winning_tool_robot = 0;
                tool_id = 0;
                for t = 1:num_tools
                    bids = tool_bids(:,t);
                    [min_tools_1, id] = min(bids);  % get lowest
                    bids(id) = [];
                    min_tools_2 = min(bids);  % get second lowest
                    dbid = min_tools_2 - min_tools_1;
                    if dbid > winning_tool_dbid
                        winning_tool_bid = min_tools_1;
                        winning_tool_dbid = dbid;
                        winning_tool_robot = id;
                        tool_id = t;
                    end
                end
                if winning_tool_bid < 0
                    winning_tool_bid = inf;
                end
            end
            
             % 3) Least contested bid (ratio)
            if resolution_type == 3
                winning_task_bid = -1;
                winning_task_dbid = 0;
                winning_task_robot = 0;
                task_id = 0;
                for t = 1:num_tasks
                    bids = task_bids(:,t);
                    [min_tasks_1, id] = min(bids);  % get lowest
                    bids(id) = [];
                    min_tasks_2 = min(bids);  % get second lowest
                    dbid = ( min_tasks_2 - min_tasks_1 ) / min_tasks_1;
                    if isnan(dbid)
                        dbid = 0;
                    end
                    if dbid > winning_task_dbid
                        winning_task_bid = min_tasks_1;
                        winning_task_robot = id;
                        winning_task_dbid = dbid;
                        task_id = t;
                    end
                end
                if winning_task_bid < 0
                    winning_task_bid = inf;
                end
                winning_tool_bid = -1;
                winning_tool_dbid = -1;
                winning_tool_robot = 0;
                tool_id = 0;
                for t = 1:num_tools
                    bids = tool_bids(:,t);
                    [min_tools_1, id] = min(bids);  % get lowest
                    bids(id) = [];
                    min_tools_2 = min(bids);  % get second lowest
                    dbid = min_tools_2 - min_tools_1;
                    if dbid > winning_tool_dbid
                        winning_tool_bid = min_tools_1;
                        winning_tool_dbid = dbid;
                        winning_tool_robot = id;
                        tool_id = t;
                    end
                end
                if winning_tool_bid < 0
                    winning_tool_bid = inf;
                end
            end
            
            % 2) Fewest bids first:
            if resolution_type == 4
                
                % EXAMPLE:
                % C = [5   6 inf 4
                %      inf 7   1 2
                %      inf 3 inf 3] 
                infs = task_bids==inf;        
                % infs = [0 0 1 0 
                %         1 0 0 0
                %         1 0 1 0]
                infs_per_task = sum(infs,1);    
                % infs_per_task = [2 0 2 0]
                bidders_per_task = size(task_bids,1) - infs_per_task;
                % bidders_per_task = [1 3 1 3]
                min_bidders = min(bidders_per_task);
                % min_bidders = 1
                cols = find(bidders_per_task == min_bidders);
                % cols = [1 3]
                [min_bids,rows] = min( task_bids(:,cols), [], 1 );
                % min_bids = [5 1]       rows = [1 2]
                [min_bid,min_bid_index] = min( min_bids );
                % min_bid = 1     min_bid_index = 2
                col = cols(min_bid_index);
                row = rows(min_bid_index);
                % col = 3     row = 2
                
                % Get the task and robot id's.
                task_id  = col;
                winning_task_bid = min_bid;
                winning_task_robot = row;
                
                winning_tool_bid = inf;
                
            end
            
            
            % Option 1) Use the lowest of the winning task and tool bids.
            if tool_bid_option == 1
                if winning_tool_bid < winning_task_bid
                    is_task = false;
                    min_bid = winning_tool_bid;
                    r_id = winning_tool_robot;
                    t_index = tool_id;
                else
                    is_task = true;
                    min_bid = winning_task_bid;
                    r_id = winning_task_robot;
                    t_index = task_id;
                end
            end
            % Option 2) If any tool was bid on, it wins.
            if tool_bid_option == 2
                if isinf(winning_tool_bid)
                    is_task = true;
                    min_bid = winning_task_bid;
                    r_id = winning_task_robot;
                    t_index = task_id;
                else
                    is_task = false;
                    min_bid = winning_tool_bid;
                    r_id = winning_tool_robot;
                    t_index = tool_id;
                end
            end

            
            % If the minimum bid was infinity, return nothing.
            if isinf(min_bid)
                r_id = [];
                is_task = [];
                t_index = [];
            end
        end
        
        % Calculates the amount to bid for each tool.
        % id  - [int]
        % pos - [x y], start position.
        %       [x2 y2], optional end position.
        % robot_expertise - [robot1ExpForT1 robot1ExpForT2 robot1ExpForT3
        %                    robot2ExpForT1 robot2ExpForT2 robot2ExpForT3 ]
        % tool_expertise - [tool1ExpForT1 tool1ExpForT2 tool1ExpForT3
        %                   tool2ExpForT1 tool2ExpForT2 tool2ExpForT3 ]
        % path_plan -  [id1 id2 ...] ids of tasks and tools
        % tasks_to_bid_on - [id1 id2 ...]
        % task_list - [id type x y base_cost
        %              id type x y base_cost]
        % objective -[1|2] 1 = minisum, 2 = minimax
        % tool_rack_list - [id x y cost]
        % travel_costs - [c11 c12 c13 ...  The travel cost between two
        %                 c21 c22 c23 ...] nodes. The first R entries are
        %                 robot start positions, the next Ta entries are
        %                 the tasks, the end To entries are tool racks.
        function bids = calculateToolBids( id, pos, robot_expertise, tool_expertise, path_plan, tasks_to_bid_on, task_list, objective, tool_rack_list, travel_costs )
            % Set the bids for all tools to inf.
            num_tools = size(tool_expertise, 1);
            num_robots = size(robot_expertise, 1);
            num_tasks = length(tasks_to_bid_on);
            bids = inf(1, num_tools);
            
            % Get currently equipped tool.
            equipped_tool = 10*path_plan( find(path_plan < 1, 1, 'last') );
            if isempty(equipped_tool)
                equipped_tool = 1;
            end
            
            % Calculate the current path cost. (Needed for MiniSum).
            curr_cost = RobotAlgs.calculatePathCost( id, pos, path_plan, task_list, robot_expertise, tool_expertise, tool_rack_list, travel_costs);
            
            % For each task:
            for i = 1:num_tasks
                % Find the best tool for that task.
                t = tasks_to_bid_on(i);
                type = task_list(t, 2);
                [~, tool] = max(tool_expertise(:,type));
                % If its our currently equipped tool, continue.
                if tool == equipped_tool
                    continue;
                end
                % Calculate the cost to complete the task with that tool.
                path = [path_plan tool/10 t];
                cost = RobotAlgs.calculatePathCost(id, pos, path, task_list, robot_expertise, tool_expertise, tool_rack_list, travel_costs);
                % MiniMax - bid is the cost.
                if objective == 2
                    bid = cost;
                end
                % MiniSum - bid is the EXTRA cost.
                if objective == 1
                    bid = cost - curr_cost;
                end
                % If the bid is a new minimum, update the bid for that tool.
                if bid < bids(tool)
                   bids(tool) = bid; 
                end
            end
            disp('');
        end
        
        % Calculates the amount to bid for each unallocated task.
        % id  - [int]
        % pos - [x y], start position.
        %       [x2 y2], optional end position.
        % robot_expertise - [robot1ExpForT1 robot1ExpForT2 robot1ExpForT3
        %                    robot2ExpForT1 robot2ExpForT2 robot2ExpForT3 ]
        % tool_expertise - [tool1ExpForT1 tool1ExpForT2 tool1ExpForT3
        %                   tool2ExpForT1 tool2ExpForT2 tool2ExpForT3 ]
        % path_plan -  [id1 id2 ...] ids of tasks and tools
        % tasks_to_bid_on - [id1 id2 ...]
        % task_list - [id type x y base_cost
        %              id type x y base_cost]
        % objective -[1|2] 1 = minisum, 2 = minimax
        % tool_rack_list - [id x y cost]
        % travel_costs - [c11 c12 c13 ...  The travel cost between two
        %                 c21 c22 c23 ...] nodes. The first R entries are
        %                 robot start positions, the next Ta entries are
        %                 the tasks, the end To entries are tool racks.
        function bids = calculateTaskBids( id, pos, robot_expertise, tool_expertise, path_plan, tasks_to_bid_on, task_list, objective, tool_rack_list, travel_costs)
            bids = zeros( 1, length(tasks_to_bid_on));
            % Calculate the current path cost. (Needed for MiniSum).
            curr_cost = RobotAlgs.calculatePathCost( id, pos, path_plan, task_list, robot_expertise, tool_expertise, tool_rack_list, travel_costs);
            % Compare each task.
            for i = 1:length(tasks_to_bid_on)
                t = tasks_to_bid_on(i);
                % Add the tasks to the end of our path (nearest neighbour).
%                 p = [path_plan t];
%                 cost = RobotAlgs.calculatePathCost( id, pos, p, task_list, robot_expertise, tool_expertise, tool_rack_list, travel_costs);
                % Insertion.
                [~, cost] = RobotAlgs.findShortestPath(id, pos, path_plan, [path_plan t], task_list, robot_expertise, tool_expertise, tool_rack_list, travel_costs);
                
                % MiniMax - bid is the cost.
                if objective == 2
                    bids(i) = cost;
                end
                % MiniSum - bid is the EXTRA cost.
                if objective == 1
                    bids(i) = cost - curr_cost;
                end
            end
        end
        
        % Finds the shortest path given task IDs to complete.
        % id  - [robot_id]
        % pos - [x y]  current position
        %       [x2 y2] (optional) desired end position.
        % prev_path - [task1 task2 task3] allows the heuristic to build on the previous path.
        % path_ids - [task1 task2 task3], tasks to be completed
        % task_list - [task_id type x y completed time remaining_completion
        %              task_id type x y completed time remaining_completion] 
        %                 remaining_completion is in seconds for a robot with expertise of 1.
        % robot_expertise - [robot1ExpForT1 robot1ExpForT2 robot1ExpForT3
        %                    robot2ExpForT1 robot2ExpForT2 robot2ExpForT3 ]
        % tool_expertise - [tool1ExpForT1 tool1ExpForT2 tool1ExpForT3
        %                   tool2ExpForT1 tool2ExpForT2 tool2ExpForT3 ]
        % tool_rack_list - [id x y cost]
        % travel_costs - [c11 c12 c13 ...  The travel cost between two
        %                 c21 c22 c23 ...] nodes. The first R entries are
        %                 robot start positions, the next Ta entries are
        %                 the tasks, the end To entries are tool racks.
        function [new_path, path_cost] = findShortestPath(id, pos, prev_path, path_ids, task_list, robot_expertise, tool_expertise, tool_rack_list, travel_costs)
            % If the path is size 1, just allocate it.
            if length(path_ids) == 1
               new_path = path_ids(1);
               path_cost = RobotAlgs.calculatePathCost(id, pos, new_path, task_list, robot_expertise, tool_expertise, tool_rack_list, travel_costs);                                                  
               return;
            end
            
            % If the new set of tasks are 1 bigger than the previous set of
            % tasks, we will insert the new task in somewhere.
            if ismember(prev_path, path_ids)  % all of previous pask are in the tasks
                if length(path_ids) - length(prev_path) == 1
                    % Find the task that isn't in the path.
                    task_id = path_ids(~ismember(path_ids,prev_path));
                    % Find the best place to insert it.
                    best_path_dist = inf;
                    best_path = [];
%                     for i = length(prev_path)+1     % nearest neighbour
                    for i = 1:length(prev_path)+1    % insertion
                        new_path = RobotAlgs.insertValueIntoVector(prev_path,task_id,i);
                        path_dist = RobotAlgs.calculatePathCost(id, pos, new_path, task_list, robot_expertise, tool_expertise, tool_rack_list, travel_costs);
                        if path_dist < best_path_dist
                            best_path_dist = path_dist;
                            best_path = new_path;
                        end
                    end 
                    new_path = best_path;
                    path_cost = best_path_dist;
                    return;
                end
            end
        end
            
        % Inserts a value into a vector.
        % vector - [double1 double2]
        % value - [double]
        % index - [integer]
        function new_vec = insertValueIntoVector( vector, value, index)
            assert( index<= numel(vector)+1);
            assert( index>=1);
            if index == numel(vector)+1
                new_vec = [vector value];
            else
                new_vec = [vector(1:index-1) value vector(index:end)];
            end
        end
        
        % Calculates the distance for a given set of tasks. Includes the
        % start position of the robot. If there is a second row in 'pos',
        % this is treated as the desired end position. If there is no
        % second row, there is no desired end position.
        % id  - [int]
        % pos - [x y], start position.
        %       [x2 y2], optional end position.
        % path_plan - [taskid1 taskid2 taskid3] Can include tool ids.
        % task_list - [id type x y base_cost
        %              id type x y base_cost]
        %                 remaining_completion is in seconds for a robot with expertise of 1.
        % robot_expertise - [robot1ExpForT1 robot1ExpForT2 robot1ExpForT3
        %                    robot2ExpForT1 robot2ExpForT2 robot2ExpForT3 ]
        % tool_expertise - [tool1ExpForT1 tool1ExpForT2 tool1ExpForT3
        %                   tool2ExpForT1 tool2ExpForT2 tool2ExpForT3 ]
        % tool_rack_list - [id x y cost]
        % travel_costs - [c11 c12 c13 ...  The travel cost between two
        %                 c21 c22 c23 ...] nodes. The first R entries are
        %                 robot start positions, the next Ta entries are
        %                 the tasks, the end To entries are tool racks.
        function [cost] = calculatePathCost(id, pos, path_plan, task_list, robot_expertise, tool_expertise, tool_rack_list, travel_costs)
            cost = 0;
            if isempty(path_plan)
                return;
            end
            base_task_expertise = robot_expertise(id,:);
            tool_id = 1;
            
            num_robots = size(robot_expertise, 1);
            num_tasks = size(task_list, 1);
            
            % Loop through path v2.
            tool_swapped = false;
            pos_index = id;
            for i=1:length(path_plan)
                task_id = path_plan(i);
                % For tasks.
                if task_id >= 1
                    % Calculate travel cost.
                    if tool_swapped % If we need to include a tool swap as well.
                        pos_to_rack = travel_costs(pos_index, num_robots+num_tasks+1:end)';
                        swap = tool_rack_list(:,4);
                        rack_to_task = travel_costs(num_robots+num_tasks+1:end, task_id+num_robots);
                        
                        travel_cost = min(pos_to_rack + swap + rack_to_task);
                        cost = cost + travel_cost;
                    else            % If we just go from task to task.
                        travel_cost = travel_costs(pos_index, task_id+num_robots);
                        cost = cost + travel_cost;
                    end
                    % Calculate completion cost.
                    task_expertise = base_task_expertise + tool_expertise(tool_id,:);
                    expertise = task_expertise( task_list(task_id,2) );
                    base_cost = task_list(task_id, 5);
                    completion_cost = base_cost / expertise;
                    cost = cost + completion_cost;
                    % Update position.
                    pos_index = task_id+num_robots;
                    tool_swapped = false;
                end
                % For tool change then task.
                if task_id < 1
                    tool_id = task_id * 10;
                    tool_swapped = true;
                    
%                     % Calculate travel cost.
%                     travel_cost = travel_costs(pos_index, tool_id+num_robots+num_tasks);
%                     cost = cost + travel_cost;
%                     swap_costs = travel_costs(pos_index, num_robots+num_tasks+1:end)' + tool_rack_list(:,4);
%                     [swap_cost, ind] = min(swap_costs);

                    % Update position.
%                     pos_index = tool_id+num_robots+num_tasks;
                end
            end
            % Return home.
            if size(pos, 1) > 1
                travel_cost = travel_costs(pos_index, id);
                if tool_id ~= 1  % swap back to no tool
                    pos_to_rack = travel_costs(pos_index, num_robots+num_tasks+1:end)';
                    swap = tool_rack_list(:,4);
                    rack_to_task = travel_costs(num_robots+num_tasks+1:end, id);

                    travel_cost = min(pos_to_rack + swap + rack_to_task);
                end
                cost = cost + travel_cost;
            end
            % Convert NaN to Inf.
            if isnan(cost)
                cost = inf;
            end
            
            % END LOOP V2
            
%             end_pos = [];
%             if size(pos, 1) > 1
%                 end_pos = pos(2,:);
%             end
%             start_pos = pos(1,:);  
%             % Loop through the path.
%             pos = start_pos;
%             for i=1:length(path_plan)
%                 task_id = path_plan(i);
%                 % For tasks:
%                 if task_id >= 1
%                     % Calculate travel cost.
%                     task_pos = task_list(task_id, 3:4);
%                     dx = task_pos(1) - pos(1);
%                     dy = task_pos(2) - pos(2);
%                     travel_cost = sqrt( dx^2 + dy^2 );
%                     cost = cost + travel_cost;
%                     % Calculate completion cost.
%                     expertise = task_expertise( task_list(task_id,2) );
%                     base_cost = task_list(task_id, 5);
%                     completion_cost = base_cost / expertise;
%                     cost = cost + completion_cost;
%                     % Update position.
%                     pos = task_pos;
%                 end
%                 % For tool changes:
%                 if task_id < 1
%                     tool_id = task_id * 10;
%                     task_expertise = base_task_expertise + tool_expertise(tool_id,:);
%                     % Calculate travel cost.
%                     tool_pos = tool_rack_list(:,2:3);
%                     dx = tool_pos(:,1) - pos(1);
%                     dy = tool_pos(:,2) - pos(2);
%                     swap_costs = sqrt( dx.^2 + dy.^2 ) + tool_rack_list(:,4);
%                     [swap_cost, ind] = min(swap_costs);
%                     cost = cost + swap_cost;
%                     % Update position.
%                     pos = tool_pos(ind,:);
%                 end
%             end
%             % Return home.
%             if ~isempty(end_pos)
%                 dx = end_pos(1) - pos(1);
%                 dy = end_pos(2) - pos(2);
%                 travel_cost = sqrt( dx^2 + dy^2 );
%                 cost = cost + travel_cost;
%             end
%             % Convert NaN to Inf.
%             if isnan(cost)
%                 cost = inf;
%             end
        end
        
        function path = removeTrailingToolSwaps( path_plan )
           ind = find(path_plan >= 1, 1, 'last');
           path = path_plan(1:ind);
        end
        
        % From wikipedia.
        function new_path = twoOptSwap(path, p1, p2)
            new_path = path;
            new_path(p1:p2) = new_path(p2:-1:p1);
%             
%             % Add the first few in order.
%             for i=1:p1-1
%                 new_path(i) = path(i);
%             end
%             % Add the nodes between the edges in reverse order.
%             j = p2;
%             for i=p1:p2
%                 new_path(i) = path(j);
%                 j = j - 1;
%             end
%             % Add the rest in order.
%             for i = p2+1:length(path)
%                 new_path(i) = path(i);
%             end
        end
    end
end



