%% Nick Sullivan 2017, The University of Adelaide


classdef State < handle  % Inheriting from 'handle' makes this return a pointer.
    properties
        %         master_state         % [bool] the 'purest' state is set to true. This is when it only swaps if absolutely
        % necessary, and is the benchmark from which all others are compared.
        %         main_state           % [bool] the state that is what would occur if we follow the decision making system.
        % These vars should not change once auctioning has begun.
        robot_list           % [id x y
                             %  id x y]      start positions of each robot
        task_list            % [id type x y base_cost
                             %  id type x y base_cost]  types, positions, and costs to complete for each task.
        tool_rack_list       % [id x y cost
                             %  id x y cost] locations and costs of tool racks.
        robot_expertise      % [expforT1 expforT2 expforT3
                             %  expforT1 expforT2 expforT3] robot's base (no tools) expertise for each task type
        tool_expertise       % [expforT1 expforT2 expforT3
                             %  expforT1 expforT2 expforT3] tool's bonus expertise for each task type
        return_home          % Whether or not the robots must return to their starting point.
        
        travel_costs         % [c11 c12 c13 c14 ...  the cost matrix to travel from one node to another
                             %  c21 c22 c23 c24 ...]  The first R entries are
                             %                 robot start positions, the next Ta entries are
                             %                 the tasks, the end To entries are tool racks.
        
        transformed          % [bool]                 
        chosen_racks         % same size as 'travel_costs'. Used in transformation, stores the best rack between two tasks/robots
        best_tools           % [bestToolForT1 bestToolForT2 ...]
        
        % These vars will change from auction round to auction round.
        state_index          % [int] the current index for the following:
        path_plan            %  {}{[id1 id2 ...]  id's of tasks and tools, in the order this robot will visit them.
                             %  [id1 id2 ...]}  stores previous path_plans
                             %  as well
                             % e.g. path_plan{state_index}{1} = [1 4 6] is the most recent path_plan (state_index) for robot 1, which
                             %      involves completing tasks 1, 4, 6.
                             % e.g. path_plan{1}{3} = [0.1] is the first path plan (1) for robot 3, which involves
                             %      swapping to tool 1.
        allocated_tasks      % {}[id1 id2 ...] id's of tasks and tools, in the order they have been allocated.
                             % e.g. allocated_tasks{4} = [1 0.2 2 0.2 4] is the tools and tasks allocated at
                             %      timestamp 4, in order.
        unallocated_tasks    % {}[id1 id2 ...] id's of tasks remaining.
        
        % Robot ID's match the row number in robot_list. Task ID's match
        % the row number in task_list. Tool ID's are 0.#, where # is the
        % ID.
        
        % Post-processed values. Will remain if state reverted.
        %         code
        minimax_cost
        minisum_cost
        %         bench_minimax_cost  % [cost1 cost2 ...] the cost if we had not swapped, one entry for each time we swap tools
        %         bench_minisum_cost  % [cost1 cost2 ...] (for the time being we only care about the last value)
        %
        %         % MACHINE LEARNING
        %         % Input variables (predictors).
        %         swap_cost       % The cost of travelling to the tool rack
        %         stay_cost       % The lowest bid we have for a task
        %         tasks_left      % The number of tasks left that we can complete
        %         swap_evenness   % How well distributed robots are to the task requirements if we swap
        %         stay_evenness   % How well distributed robots are to the task requirements if we stay
        %
        %         % Output variables (reward).
        %         reward          % The benchmark path cost divided by the path cost
        
        
    end
    methods
        % Constructor.
        function obj = State()
            obj.transformed = false;
        end
        
        % Populates the 'travel_costs' property using euclidean distance.
        % travel_costs - [c11 c12 c13 ...  The travel cost between two
        %                 c21 c22 c23 ...] nodes. The first R entries are
        %                 robot start positions, the next Ta entries are
        %                 the tasks, the end To entries are tool racks.
        function populateTravelCosts(s)
            num_robots = size(s.robot_list, 1);
            num_tasks = size(s.task_list, 1);
            num_tool_racks = size(s.tool_rack_list, 1);
            num_entries = num_robots+num_tasks;
            
            s.travel_costs = zeros(num_entries);
            x = [s.robot_list(:,2); s.task_list(:,3)];
            y = [s.robot_list(:,3); s.task_list(:,4)];
            if ~isempty(s.tool_rack_list)
                num_entries = num_entries + num_tool_racks;
                x = [x; s.tool_rack_list(:,2)];
                y = [y; s.tool_rack_list(:,3)];
            end
            
            for i=1:num_entries
               for j=1:num_entries
                   dx = x(j) - x(i);
                   dy = y(j) - y(i);
                   dist = sqrt( dx^2 + dy^2 );
                   s.travel_costs(i, j) = dist;
               end
            end
        end
        
        % Calculates minimax and minisum costs for the last state.
        function calculateMiniCosts(this)
            s = this;
            num_robots = size(s.robot_list,1);
            
            s.minisum_cost = 0;
            s.minimax_cost = 0;
            for r = 1:num_robots
                pos = s.robot_list(r,2:3);      % Add entry for return home.
                if s.return_home
                    pos = [pos; pos];
                end
                cost = RobotAlgs.calculatePathCost(r, pos, s.path_plan{end}{r}, s.task_list, s.robot_expertise, s.tool_expertise, s.tool_rack_list, s.travel_costs);
                s.minisum_cost = s.minisum_cost + cost;
                if cost > s.minimax_cost
                    s.minimax_cost = cost;
                end
            end
            
        end
        
        % Populates the machine learning input variables for the given
        % state index.
        function predictors = calculateMLPredictors(this, r_id, state_index)
            s = this;
            si = state_index;
            num_robots = size(s.robot_list, 1);
            % 1) swap_cost. Calculate distance to tool rack.
            pos = s.robot_list(r_id, 2:3);
            bid = RobotAlgs.calculateToolBid( r_id, pos, s.robot_expertise, s.tool_expertise, s.path_plan{si}{r_id}, 1, s.task_list, s.tool_rack_list);
            s.swap_cost = bid;
            % 2) stay_cost. Re-calculate bids, and find the minimum.
            bids = RobotAlgs.calculateTaskBids( r_id, pos, s.robot_expertise, s.tool_expertise, s.path_plan{si}{r_id}, s.unallocated_tasks{si}, s.task_list, s.tool_rack_list);
            s.stay_cost = min(bids);
            % 3) tasks_left. Count the non-inf bids.
            s.tasks_left = sum(~isinf(bids));
            % 4) swap_evenness and 5) stay_evenness. Calculate the evenness metric
            % for all tool values.
            equipped_tools = ones(num_robots, 1);
            for r = 1:num_robots
                entry = find( s.path_plan{si}{r} < 1, 1, 'last');
                if ~isempty(entry)
                    tool = s.path_plan{si}{r}(entry)*10;
                    equipped_tools(r) = tool;
                end
            end
            em = RobotAlgs.calculateEvennessMetric( r_id, s.robot_expertise, s.tool_expertise, equipped_tools, s.unallocated_tasks{si}, s.task_list);
            s.stay_evenness = em( equipped_tools(r_id), : );
            em( equipped_tools(r_id), : ) = [];
            % Remove all those with greater than minimum infs.
            min_infs = min(em(:,1));
            em( em(:,1) > min_infs, : ) = [];
            % Select the one with the minimum spread.
            [~,index] = min(em(:,2));
            s.swap_evenness = em(index, :);
            
            predictors = [s.swap_cost s.stay_cost s.tasks_left s.swap_evenness s.stay_evenness s.reward];
        end
        
        % Make a copy of a handle object.
        function new = copy(this)
            % Instantiate new object of the same class.
            new = State();
            % Copy all non-hidden properties.
            p = properties(this);
            for i = 1:length(p)
                new.(p{i}) = this.(p{i});
            end
        end
        
        % Plots the state.
        % objective - [1|2] 1 = minisum, 2 = minimax
        function plotState(this, state_index, objective)
            thick_mode = true;  % change this if you want
            if thick_mode
                task_size = 40;
                tool_size = 120;
                robot_size = 140;
                line_width = 2;
            else
                task_size = 25;
                tool_size = 50;
                robot_size = 40;
                line_width = 1;
            end
            clf;
            hold on;
            num_robots = size(this.robot_list, 1);
            num_tool_racks = size(this.tool_rack_list, 1);
            colours = lines(num_robots+1);
            num_allocations = length(this.allocated_tasks{end});
            if state_index > 0
                pid = state_index;
            else
                pid = num_allocations+1;
            end

            % Plot robots.
            for r = 1:num_robots
                x = zeros(length(this.path_plan{pid}{r})+1, 1);
                y = x;
                x(1) = this.robot_list(r,2);
                y(1) = this.robot_list(r,3);
                already_home = false;
                has_a_tool = false;
                for i = 1:length(this.path_plan{pid}{r})
                    t_id = this.path_plan{pid}{r}(i);
                    if t_id < 1  % is a tool
                        has_a_tool = true;
%                         x(i+1) = 5;      % HARD CODED
%                         y(i+1) = 4;
                        % dirty code here, assumes the tool rack we want is
                        % the best one for that task.
                        start_x = x(i);
                        start_y = y(i);
                        if i == length(this.path_plan{pid}{r})
                            disp('tool swapping home');
                            end_x = this.robot_list(r,2);
                            end_y = this.robot_list(r,3);
                            tool_pos = this.tool_rack_list(:,2:3);
                            dx1 = tool_pos(:,1) - start_x;
                            dy1 = tool_pos(:,2) - start_y;
                            dx2 = tool_pos(:,1) - end_x;
                            dy2 = tool_pos(:,2) - end_y;
                            swap_costs = sqrt( dx1.^2 + dy1.^2 ) + sqrt( dx2.^2 + dy2.^2 ) + this.tool_rack_list(:,4);
                            [swap_cost, ind] = min(swap_costs);
                            x(i+1) = tool_pos(ind,1);
                            y(i+1) = tool_pos(ind,2);
                            x = [x; this.robot_list(r,2)];
                            y = [y; this.robot_list(r,3)];
                            already_home = true;
%                             x = x(1:i);
%                             y = y(1:i);
                            continue;
                        end
                        next_t_id = this.path_plan{pid}{r}(i+1);
                        while next_t_id < 1
                            disp('Double tool swap');
                            i = i+1;
                            next_t_id = this.path_plan{pid}{r}(i+1);
                        end
                        end_x = this.task_list(next_t_id, 3);
                        end_y = this.task_list(next_t_id, 4);
                        % calc dists
                        tool_pos = this.tool_rack_list(:,2:3);
                        dx1 = tool_pos(:,1) - start_x;
                        dy1 = tool_pos(:,2) - start_y;
                        dx2 = tool_pos(:,1) - end_x;
                        dy2 = tool_pos(:,2) - end_y;
                        swap_costs = sqrt( dx1.^2 + dy1.^2 ) + sqrt( dx2.^2 + dy2.^2 ) + this.tool_rack_list(:,4);
                        [swap_cost, ind] = min(swap_costs);
                        x(i+1) = tool_pos(ind,1);
                        y(i+1) = tool_pos(ind,2);
                    else
                        x(i+1) = this.task_list(t_id, 3);
                        y(i+1) = this.task_list(t_id, 4);
                    end
                end
                if this.return_home && ~already_home
                    if has_a_tool
                        start_x = x(end);
                        start_y = y(end);
                        end_x = this.robot_list(r,2);
                        end_y = this.robot_list(r,3);
                        tool_pos = this.tool_rack_list(:,2:3);
                        dx1 = tool_pos(:,1) - start_x;
                        dy1 = tool_pos(:,2) - start_y;
                        dx2 = tool_pos(:,1) - end_x;
                        dy2 = tool_pos(:,2) - end_y;
                        swap_costs = sqrt( dx1.^2 + dy1.^2 ) + sqrt( dx2.^2 + dy2.^2 ) + this.tool_rack_list(:,4);
                        [swap_cost, ind] = min(swap_costs);
                        x = [x; tool_pos(ind,1)];
                        y = [y; tool_pos(ind,2)];
                    end
                    x = [x; this.robot_list(r,2)];
                    y = [y; this.robot_list(r,3)];
                end
                plot(x, y, 'Color', colours(r,:), 'Linewidth', line_width);
                scatter(x(1), y(1), robot_size, 'MarkerFaceColor', colours(r,:), 'MarkerEdgeColor', [0 0 0], 'Linewidth', line_width);
            end
                        
            % Plot tasks.
            tasks_by_colour = true;
            if tasks_by_colour
                for i = 1:max(this.task_list(:,2))
                    x = this.task_list(:, 3);
                    y = this.task_list(:, 4);
                    x = x( this.task_list(:,2) == i );
                    y = y( this.task_list(:,2) == i );
%                     scatter(x, y, task_size, 'filled', 'MarkerEdgeColor',[0 0 0], 'MarkerFaceColor',[0 0 0]);
%                     if i == 1
%                         scatter(x, y, task_size, 'filled', 'MarkerEdgeColor',[0 0 0], 'MarkerFaceColor',[0 0 0]);
%                     else
                        scatter(x, y, task_size, 'filled', 'MarkerEdgeColor',[0 0 0]);% 'MarkerFaceColor',colours(i,:));
%                     end
                        
                end
            else
                scatter_types = ['*';'+';'d'];
                for i = 1:max(this.task_list(:,2))
                    x = this.task_list(:, 3);
                    y = this.task_list(:, 4);
                    x = x( this.task_list(:,2) == i );
                    y = y( this.task_list(:,2) == i );
                    scatter(x, y, task_size, scatter_types(i), 'MarkerEdgeColor',[0 0 0], 'MarkerFaceColor',[0 0 0],'Linewidth', 1);
                end
            end
            
            % Plot the tool change.
            if ~isempty(this.tool_rack_list)
                x = this.tool_rack_list(:,2);
                y = this.tool_rack_list(:,3);
                scatter(x, y, tool_size, 'square', 'MarkerFaceColor', [0 0 0], 'MarkerEdgeColor', [0 0 0]);
            end
            
%             xlim([0 100]);
%             ylim([0 100]);
            if objective == 1
                title_str = ['MiniSum: ' num2str(this.minisum_cost)];
            end
            if objective == 2
                title_str = ['MiniMax: ' num2str(this.minimax_cost)];
            end
            title(title_str);
%             xlabel(['order: ' num2str(this.allocated_tasks{end})]);
            
        end
        
        % Represents the current state as a gene. Must be completed.
        function g = convertToGASolution(s)
            g = GASolution(s.robot_list, s.task_list, s.tool_rack_list, ...
                s.robot_expertise, s.tool_expertise, s.return_home, 2, false);
            paths = s.path_plan{s.state_index};
            task_paths = cell(g.num_robots,1);
            tool_paths = cell(g.num_robots,1);
            for r = 1:g.num_robots
                % Convert the state path [1 4 3 0.2 5 0.1 6] to
                % task path [1 4 3 5 6] and
                % tool path [1 1 1 2 1].
                task_path = zeros(1, g.num_tasks);
                tool_path = zeros(1, g.num_tasks);
                current_tool = 1;
                index = 0;
                path = paths{r};
                len = length(path);
                for t = 1:len
                    t_id = path(t);
                    % If its a tool, update the tool we're on.
                    if t_id < 1
                        current_tool = t_id * 10;
                        continue;
                    end
                    % If it's a task, log it.
                    index = index + 1;
                    task_path(index) = t_id;
                    tool_path(index) = current_tool;
                end
                task_path = task_path(1:index);
                tool_path = tool_path(1:index);
                
                task_paths{r} = task_path;
                tool_paths{r} = tool_path;
            end
            g.setPaths(task_paths,tool_paths);
        end
        
        function exportScenario(s, file_name)
            % Write the scenario file (num markers, num robots, func per
            % robot), then vehicle start positions and target locations.
            fid = fopen([file_name],'wt');
            fprintf(fid, [num2str(size(s.task_list,1)) '\n']);
            fprintf(fid, [num2str(size(s.robot_list,1)) '\n']);
            fprintf(fid, '\n');
            for r = 1:size(s.robot_list,1)
                fprintf(fid, '0\n');  % no functional heterogeniety
            end

            fprintf(fid, '\n');
            for r = 1:size(s.robot_list,1) 
                fprintf(fid, [num2str(s.robot_list(r,2)) ' ' num2str(s.robot_list(r,3)) '\n']);
            end
            fprintf(fid, '\n');
            for t = 1:size(s.task_list,1)
                fprintf(fid, [num2str(s.task_list(t,3)) ' ' num2str(s.task_list(t,4)) '\n']);
            end
        end
        
        function exportCosts(s, file_name)
           % Write the cost files for each robot, in the form:
           %     t1 t2 t3 ... depot
           % t1
           % t2
           % t3
           % ...
           % depot
           num_robots = size(s.robot_list,1);
           num_tasks = size(s.task_list,1);
           for r = 1:size(s.robot_list,1)
               task_edges = s.travel_costs(num_robots+1:num_robots+num_tasks, num_robots+1:num_robots+num_tasks);
               depot_edges = s.travel_costs(r, num_robots+1:num_robots+num_tasks);
               
               costs = [task_edges; depot_edges];
               costs = [costs [depot_edges'; 0]];
               
               file_name2 = [file_name '-' num2str(r)];
               dlmwrite(file_name2, costs, 'delimiter', '\t', 'precision', '%.2f');
           end
        end
                
    end
    
    methods(Static)
        
        
        % Converts a state (with tools and tool racks) into a state that
        % has no tools or tool racks, but adjusts the costs accordingly.
        function [state, E_lim, E_max] = transformToMTSP(s)
            num_robots = size(s.robot_list,1);
            num_tasks = size(s.task_list, 1);
            num_tools = size(s.tool_expertise,1);
            
            % 1) Select the best tools for each task and the starting tools.
            s.chosen_racks = zeros(num_robots+num_tasks);
            [best_exp, best_tools] = max(s.tool_expertise, [], 1);
            s.best_tools = best_tools;
            
%             best_tools = [zeros(1, num_robots) best_tools];
            % 2) Transform edges.
            biggest_swap_cost = 0;
            for i = 1:num_robots+num_tasks
                best_tool_i = 0;
                if i > num_robots
                    task_type = s.task_list(i-num_robots,2);
                    best_tool_i = best_tools(task_type);
                end
                
                for j = 1:num_tasks 
                    task_type = s.task_list(j,2);
                    if best_tool_i == best_tools(task_type) % same tool = do nothing
                        continue;
                    end
                    i_to_rack = s.travel_costs(i, num_robots+num_tasks+1:end)';
                    swap = s.tool_rack_list(:,4);
                    rack_to_j = s.travel_costs(num_robots+num_tasks+1:end, j+num_robots);
                    [travel_cost, rack_id] = min(i_to_rack + swap + rack_to_j);
                    s.travel_costs(i, j+num_robots) = travel_cost;
                    s.chosen_racks(i, j+num_robots) = rack_id;
                    if travel_cost > biggest_swap_cost
                       biggest_swap_cost = travel_cost; 
                    end
                    % make symmetrical.
                    s.travel_costs(j+num_robots, i) = travel_cost;
                    s.chosen_racks(j+num_robots, i) = rack_id;
                end
            end
            
            % Calculate loss of optimality.
            C = s.task_list(1,5);
            E_lim = C/(C + biggest_swap_cost*2);
            temp_exp = s.tool_expertise;
            for i=1:size(temp_exp,1)
               temp_exp(i+1,i) = 0; 
            end
            E_max = max(max(temp_exp));
            disp(['Limit for optimality: ' num2str(E_lim)]);
            disp(['Actual value: ' num2str(E_max)]);
            
            
            % 3) Set all tool expertise to 0.
            s.tool_expertise = zeros(size(s.tool_expertise) );
            % 4) Add best tool expertise to all robots.
            s.robot_expertise = s.robot_expertise + best_exp;
            
            s.transformed = true;
            state = s; 
            

        end
        
        function state = transformFromMTSP(s)
            if ~s.transformed
                state = s;
                return;
            end
            num_robots = size(s.robot_list,1);
            num_tasks = size(s.task_list, 1);
            num_tools = size(s.tool_expertise,1);
            % Go through path.
            for r = 1:num_robots
                path = s.path_plan{s.state_index}{r};
                new_path = path;
                num_tools_added = 0;
                prev = r;
                for i = 1:length(path)
                    curr = path(i);
                    rack = s.chosen_racks(prev, curr+num_robots);
                    if rack ~= 0
                        tool = s.best_tools( s.task_list(curr,2) );
                        new_path = RobotAlgs.insertValueIntoVector( new_path, tool/10, i+num_tools_added );
                        num_tools_added = num_tools_added + 1;
                    end
                    prev = curr+num_robots;
                end
                if ~isempty(path) && s.return_home
                   rack = s.chosen_racks(path(end)+num_robots, r);
                   tool = 0.12345; % any other tool
                   new_path = [new_path tool];
                end
                s.path_plan{s.state_index}{r} = new_path;
            end
            state = s;
        end
        
        % state_list - {} cell array of states
        % objective - [1|2] 1: minisum, 2: minimax
        % tool_bid_option - [1|2]  1:  tool bids are the cost of tool swap + completing one task
        %                          2:  tool bids are binary and allocated first - it is up to the robot to want a tool
        % display - [true|false] whether to display a graph during
        function fin_state_list = solveStates(state_list, objective, tool_bid_option, display)
            
            fin_state_list = [];
            fig_counter = 1;
            while ~isempty(state_list)
                % Pop the state.
                s = state_list(1);
                state_list(1) = [];
                
                % Get some states vars.
                si = s.state_index;
                num_robots = size(s.robot_list, 1);
                num_unallocated_tasks = length(s.unallocated_tasks{si});
                num_allocations = length(s.allocated_tasks{si});
                num_tools = size(s.tool_expertise, 1);
                equipped_tools = ones(num_robots, 1);
                for r = 1:num_robots
                    entry = find( s.path_plan{si}{r} < 1, 1, 'last');
                    if ~isempty(entry)
                        tool = s.path_plan{si}{r}(entry)*10;
                        equipped_tools(r) = tool;
                    end
                end
                
                % Calculate all task bids.
                task_bids = zeros(num_robots, num_unallocated_tasks);
                for r = 1:num_robots
                    pos = s.robot_list(r,2:3);      % Add entry for return home.
                    if s.return_home
                        pos = [pos; pos];
                    end
                    task_bids(r,:) = RobotAlgs.calculateTaskBids( r, pos, s.robot_expertise, s.tool_expertise, s.path_plan{si}{r}, s.unallocated_tasks{si}, s.task_list, objective, s.tool_rack_list, s.travel_costs);
                end
                
                % Tool bid option 1) bid on all tools every round, the bid
                % is the lowest cost that a robot can both tool-swap and
                % complete a task.
                if tool_bid_option == 1
                    tool_bids = inf(num_robots, num_tools);
                    for r = 1:num_robots
                        pos = s.robot_list(r,2:3);      % Add entry for return home.
                        if s.return_home
                            pos = [pos; pos];
                        end
                    
                        tool_bids(r,:) = RobotAlgs.calculateToolBids( r, pos, s.robot_expertise, s.tool_expertise, s.path_plan{si}{r}, s.unallocated_tasks{si}, s.task_list, objective, s.tool_rack_list, s.travel_costs );
                    end
                end
                
                % Tool bid option 2) determine if any robots want to swap,
                % and make is so tool bids are allocated first.
                if tool_bid_option == 2
                    tool_bids = inf(num_robots, num_tools);
                    for r = 1:num_robots
                        desired_tool = RobotAlgs.selectBestTool( r, equipped_tools, s );
                        if desired_tool ~= equipped_tools(r)
                            tool_bids(r, desired_tool) = 1;
                            break;
                        end
                    end
                end

                % Calculate winner and what is allocated. Tools will be
                % allocated first, if possible.
                [r_id, is_task, t_index] = RobotAlgs.auctionDecide(task_bids, tool_bids, tool_bid_option); % t_index is the index of unallocated_tasks that won, or the row of tool info
                
                % If we've gotten nothing back, it means nothing was allocated. So we
                % should end the state here.
                if isempty(r_id)
                    disp('Check this!');
                    fin_state_list = [fin_state_list s];
                    unused_storage_indices = cellfun('isempty',s.path_plan);
                    s.path_plan(unused_storage_indices) = []; % remove unused path_plan storage
                    s.allocated_tasks(unused_storage_indices) = [];
                    s.unallocated_tasks(unused_storage_indices) = [];
                    continue;
                end
                
                % Allocate a tool if that's what won.
                if ~is_task
                    % Update path of the winner.
                    tool_id = t_index / 10;
                    s.state_index = si+1;
                    s.path_plan{si+1} = s.path_plan{si};
                    s.path_plan{si+1}{r_id} = [s.path_plan{si}{r_id} tool_id];
                    s.unallocated_tasks{si+1} = s.unallocated_tasks{si};
                    s.allocated_tasks{si+1} = [s.allocated_tasks{si} tool_id];
                    % Add to processing state list.
                    state_list = [state_list s];
                    continue;
                end
                
                % Or allocate a task.
                task_id = s.unallocated_tasks{si}(t_index);          % get task
                s.state_index = si+1;
                s.path_plan{si+1} = s.path_plan{si};
                pos = s.robot_list(r_id,2:3);
                if s.return_home
                    pos = [pos; pos];
                end
                s.path_plan{si+1}{r_id} = RobotAlgs.findShortestPath(r_id, pos, s.path_plan{si}{r_id}, [s.path_plan{si}{r_id} task_id], s.task_list, s.robot_expertise, s.tool_expertise, s.tool_rack_list, s.travel_costs);
                
                s.unallocated_tasks{si+1} = s.unallocated_tasks{si};
                s.unallocated_tasks{si+1}(t_index) = [];               % remove from unallocated
                s.allocated_tasks{si+1} = [s.allocated_tasks{si} task_id]; % add to allocated
                
                % If all tasks are complete, add to finished state list. If tasks
                % remain, add to processing state list.
                if isempty(s.unallocated_tasks{si+1})
                    % Remove any trailing tool swaps from paths.
                    for r = 1:num_robots
                        s.path_plan{si+1}{r} = RobotAlgs.removeTrailingToolSwaps(s.path_plan{si+1}{r});
                    end
                    
                    fin_state_list = [fin_state_list s];
                    unused_storage_indices = cellfun('isempty',s.path_plan);
                    s.path_plan(unused_storage_indices) = []; % remove unused path_plan storage
                    s.allocated_tasks(unused_storage_indices) = [];
                    s.unallocated_tasks(unused_storage_indices) = [];
                    s.calculateMiniCosts();
                    
                else
                    state_list = [state_list s];
                end
                
                if display
                    s.plotState(s.state_index, objective);
                    pause(0.01);
%                     fig = gcf;
%                     fig.PaperPositionMode = 'auto';
%                     fig_pos = fig.PaperPosition;
%                     fig.PaperSize = [fig_pos(3) fig_pos(4)];
%                     name = ['Auction' num2str(fig_counter)];
%                     print(fig, '-dpdf', '-r150', name);
%                     fig_counter = fig_counter + 1;
                end
            end
            
        end
        
        
        function s = perform2Opt(s)
            % 2-opt improvement: "Comparison of Approximate Approaches to Solving the
            % Travelling Salesman Problem and its Application to UAV Swarming"
            num_robots = size(s.robot_list, 1);
            for r = 1:num_robots
                improved = true;
                % Set start and end position.
                pos = s.robot_list(r, 2:3);
                if s.return_home
                    pos = [pos; pos];
                end
                path = s.path_plan{end}{r};
                % Keep doing swaps until its no longer improving.
                while improved
                    improved = false;
                    best_distance = RobotAlgs.calculatePathCost(r, pos, path, s.task_list, s.robot_expertise, s.tool_expertise, s.tool_rack_list, s.travel_costs);
                    % Pick a start node.
                    for i=1:length(path)-1
                        % Pick an end node.
                        for k=i+1:length(path)
                            % Swap.
                            new_path = RobotAlgs.twoOptSwap(path, i, k);
                            new_distance = RobotAlgs.calculatePathCost(r, pos, new_path, s.task_list, s.robot_expertise, s.tool_expertise, s.tool_rack_list, s.travel_costs);
                            if (new_distance < best_distance)
                                path = new_path;
                                s.path_plan{end}{r} = new_path;
                                
                                improved = true;
                            end
                        end
                    end
                end
            end
            s.calculateMiniCosts();
        end
        
    end
end



