%% Nick Sullivan 2017, The University of Adelaide
% This function performs a sequential single-item auction to solve a
% multiple travelling salesman problem where tasks require tools. 
function [result, result_with_2_opt] = performAuction( scenario_num, num_robots, sec_best_exp, num_tasks_per_type, num_task_types, num_tool_racks, tsplib_name )
%% USER OPTIONS
display_graphs = true;
incremental_plot = false;
objective = 1;          % 1 - minisum
                        % 2 - minimax
return_home = true;     % if robots have to return to their start point.
tool_bid_option = 1;    % 1:  tool bids are the cost of tool swap + completing one task
                        % 2:  tool bids are binary and allocated first - it is
                        %     up to the robot to want a tool
using_tools = true;
transform = true;       % if the system should be transform to a standard MTSP first
preallocate = 500;      % prevents memory resizing


if nargin < 2           % Optional arguments
    num_robots = 3;   
end
if nargin < 3
    sec_best_exp = 0.0;
end
if nargin < 4
    num_tasks_per_type = 10;
end
if nargin < 5
    num_task_types = 5;   % also number of tools
end
if nargin < 6
   num_tool_racks = 4;
end
if nargin < 7
    use_tsplib = false;
else
    use_tsplib = true;
end

%% Set the scenario
addpath Simulator
addpath TSPLib
rng(scenario_num);
s = State();
s.return_home = return_home;

s.robot_list = [[1:num_robots]' rand(num_robots,2)*100];
task_num = 1;
s.task_list = [];
for i = 1:num_task_types
%     if i==2
%         num_tasks_per_type = 1;
%     end
    for j = 1:num_tasks_per_type
        task = [task_num i 100*rand 100*rand 10];
        task_num = task_num + 1;
        s.task_list = [s.task_list; task];
    end
end

s.state_index = 1;
s.allocated_tasks    = cell(preallocate, 1);
s.allocated_tasks{1} = [];
s.unallocated_tasks    = cell(preallocate, 1);
s.unallocated_tasks{1} = s.task_list(:,1);
s.path_plan    = cell(preallocate, 1);
s.path_plan{1} = cell( size(s.robot_list, 1), 1 );   % empty path plans

if using_tools
%     tools = zeros(num_task_types);  % COMMENT OUT UNUSED ONES
    % tools = rand(num_task_types); % one or the other
    tools = sec_best_exp*ones(num_task_types);
    
    tools = max(tools, eye(num_task_types));
    
    s.tool_expertise = [zeros(1, num_task_types); tools];
    s.robot_expertise = zeros(num_robots, num_task_types);  % use tools
    
    s.tool_rack_list = [ ones(num_tool_racks,1), rand(num_tool_racks,2)*100, 10*ones(num_tool_racks,1) ];
    
%     s.tool_rack_list = [1 20 20 10  % [id x y cost] the location of the tool racks
%                         2 80 20 10
%                         3 80 80 10
%                         4 20 80 10];
else
    s.tool_expertise = zeros(1,num_task_types);
    s.robot_expertise = ones(num_robots, num_task_types); % use innate xp
%     s.robot_expertise = [1 1 0 0
%                          1 0 1 0
%                          1 0 0 1];
end

% OVERRIDE, ADDING OWN STATES.
if use_tsplib
    s = loadStateFromTSPLIB(['TSPLib/' tsplib_name '.tsp'], num_robots, num_tool_racks, num_task_types);
end
% END OVERRIDE.

s.populateTravelCosts();

% Transform the state into a standard M-TSP problem:
if transform
    [s, E_lim, E_max] = State.transformToMTSP(s);
end
state_list = s;


%% Solve the state
fin_state_list = State.solveStates(state_list, objective, tool_bid_option, incremental_plot);

%% Process the states to determine which ones performed best.
% disp(['state_index: ' num2str(fin_state_list.state_index) ]);
% Calculate costs.
min_cost = inf;
min_cost_state = 0;
for i = 1:length(fin_state_list)
    s = fin_state_list(i);
    s.calculateMiniCosts();
    if objective == 1
        if s.minisum_cost < min_cost
            min_cost = s.minisum_cost;
            min_cost_state = i;
        end
    end
    if objective == 2
        if s.minimax_cost < min_cost
            min_cost = s.minimax_cost;
            min_cost_state = i;
        end
    end
end
% Get the best solution.
s = fin_state_list(min_cost_state);
result = min_cost;

if display_graphs
    figure();
    s.plotState(0, objective);
    pause(0.01);
end



%% Perform 2-opt improvement for each robot
result_with_2_opt = 0;
s = State.perform2Opt(s);
if objective == 1
    result_with_2_opt = s.minisum_cost;
end
if objective == 2
    result_with_2_opt = s.minimax_cost;
end
if display_graphs
    figure();
    s.plotState(0, objective);
end


%% Convert back
if s.transformed
    s = State.transformFromMTSP(s);
    if display_graphs
        figure();
        s.plotState(0, objective);
        pause(0.01);
    end
end


end

% % % Comparing to OC.
% % task_num = 1;
% % s.task_list = [];
% % for r = 1:num_robots
% %    for i = 1:10
% %        if r ~= 4
% %            continue;
% %        end
% %        task = [task_num r 10*rand 10*rand 0];
% %        task_num = task_num + 1;
% %        s.task_list = [s.task_list; task];
% %    end
% %    for i = 1:10
% %        if r == 4
% %            continue;
% %        end
% %        task = [task_num num_robots+1 10*rand 10*rand 0];
% %        task_num = task_num + 1;
% %        s.task_list = [s.task_list; task];
% %    end
% % end
% % % for i=11:20
% % %     s.task_list(i,3) = s.task_list(i,3) + 10;
% % % end
% % % for i=21:30
% % %     s.task_list(i,3) = s.task_list(i,3) + 10;
% % %     s.task_list(i,4) = s.task_list(i,4) + 10;
% % % end
% % % for i=31:40
% % %     s.task_list(i,4) = s.task_list(i,4) + 10;
% % % end
% % 
% % s.robot_expertise = [1 0 0 0 1
% %                      0 1 0 0 1
% %                      0 0 1 0 1
% %                      0 0 0 1 1];
% % % End OC.
