%% Nick Sullivan 2017, The University of Adelaide
% This function performs a genetic algorithm to solve a
% multiple travelling salesman problem where tasks require tools. 
function [result, result_with_2_opt] = performGA( scenario_num, num_robots, sec_best_exp, num_tasks_per_type, num_iterations, num_task_types, num_tool_racks, tsplib_name)
%% USER OPTIONS
display_graphs = false;
objective = 1;          % 1 - minisum
                        % 2 - minimax
return_home = true;     % if robots have to return to their start point.

using_tools = true;
transform = true;       % if the system should be transform to a standard MTSP first (basically the task-tool pairings are locked)
tool_seed = 2;%1 + 1*transform;
auction_seed = false;    % whether to seed with auction results or to use random

max_pop_size = 100;
num_replacement = 0.2*max_pop_size;    % must be even
crossover_prob = 0.95;

if nargin < 2
    num_robots = 3;
end
if nargin < 3
    sec_best_exp = 0.2;
end
if nargin < 4
    num_tasks_per_type = 2;
end
if nargin < 5
    num_iterations = 7e3;
end
if nargin < 6
    num_task_types = 5;   % also number of tools
end
if nargin < 7
   num_tool_racks = 8; 
end
if nargin < 8
    use_tsplib = false;
else
    use_tsplib = true;
end
disp(['num iter: ' num2str(num_iterations)]);

% there are more GA options in the code
%% Set the scenario
addpath Simulator
addpath TSPLib
rng(scenario_num);
% robot_list = [1 1 1             % [id x y]
%               2 9 1
%               3 9 9];
robot_list = [[1:num_robots]' rand(num_robots,2)*100];

task_num = 1;
task_list = [];
for i = 1:num_task_types
    for j = 1:num_tasks_per_type
        task = [task_num i 100*rand 100*rand 10];
        task_num = task_num + 1;
        task_list = [task_list; task];
    end
end



if using_tools
%    tools = zeros(num_task_types);  % one or the other
%      tools = sec_best_exp*rand(num_task_types); % one or the other
    tools = sec_best_exp*ones(num_task_types);
    
    tools = max(tools, eye(num_task_types));
    tool_expertise = [zeros(1, num_task_types); tools];
    robot_expertise = zeros(num_robots, num_task_types);  % use tools
    
    tool_rack_list = [ ones(num_tool_racks,1), rand(num_tool_racks,2)*100, 10*ones(num_tool_racks,1) ];
    
%     tool_rack_list = [1 20 20 10  % [id x y cost] the location of the tool racks
%                     2 80 20 10
%                     3 80 80 10
%                     4 20 80 10];
else
    tool_expertise = zeros(1,num_task_types);
    robot_expertise = ones(num_robots, num_task_types); % use innate xp
    tool_rack_list = [];
end

% OVERRIDE, ADDING OWN STATES.
if use_tsplib
%     [robot_list, task_list, tool_rack_list, robot_expertise, tool_expertise] = loadGeneFromTSPLIB(['TSPLib/' tsplib_name '.tsp'], num_robots, num_tool_racks, num_task_types);
    [robot_list, task_list, robot_expertise] = loadGeneFromTSPLIB2(['TSPLib/' tsplib_name '.tsp'], num_robots);
end
% END OVERRIDE.
% rng('shuffle');

num_tasks = size(task_list,1);
num_tools = size(tool_expertise,1);
num_robots = size(robot_expertise, 1);
gene_len = num_tasks*2+num_robots;
%% Prepare the GA
fig_counter = 1;
% Randomly generate a number of genes.
population = cell(max_pop_size, 1);
pop_size = 0;
while pop_size < max_pop_size
    g = GASolution(robot_list, task_list, tool_rack_list, robot_expertise, tool_expertise, return_home, tool_seed, auction_seed, objective, transform);
    
    % Clone the first one if we want.
    if auction_seed
        while pop_size < max_pop_size
            pop_size = pop_size + 1;
            population{pop_size} = g;
            g = g.copy();
        end
    end
    % Check for duplicates.
    duplicate = false;
%     for i=1:pop_size
%         gene = population{i}.gene;
%         if g.gene == gene
%             duplicate = true;
%         end
%     end
    if ~duplicate
        pop_size = pop_size + 1;
        population{pop_size} = g;
    end
end

% Sort the population
if objective == 1
    costs = cellfun(@(x)x.minisum, population);
end
if objective == 2
    costs = cellfun(@(x)x.minimax, population);
end
[~,sort_index] = sort(costs);
population = population(sort_index);
best_cost = inf;
disp('Population set!');

%% Solve using GA

% Repeat for a while:
cost_tracking = zeros(num_iterations, 1);
identical_tracking = zeros(num_iterations, 1);
for iteration = 1:num_iterations
    % Display progress.
    if mod(iteration, num_iterations/20) == 0
       disp(['Iteration: ' num2str(iteration)]);
    end
    % Roulette-wheel selection.
    % Select genes using fitness probability (roulette wheel). Lower
    % cost is more likely to be picked.
%     probability = 1 - (costs ./ sum(costs));
%     probability = probability ./ sum(probability);  % flip odds and normalise
    % Rank-based roulette-wheel selection.
    ranks = pop_size:-1:1;
%     probability = ranks ./ sum(ranks);
    probability = ranks.^2 ./ sum(ranks.^2);
    
    % Perform selection
    cdf = cumsum(probability);      % forms probability brackets
    candidates = zeros(num_replacement, 1);
    for i=1:num_replacement
        U = rand;
        candidates(i) = find(cdf >= U, 1, 'first');
    end
    
    % Perform crossover to find the best new child. Or duplicate the
    % parents according to crossover probability.
    for i=1:2:num_replacement
        best1 = candidates(i  );   %best to breed
        best2 = candidates(i+1);
        worst1 = max_pop_size+1-i; %worst to kill
        worst2 = max_pop_size-i;
        
        mother = population{best1};
        father = population{best2};
        if rand() < crossover_prob  %crossover
            GAAlgs.crossover(mother, father, population{worst1});
            GAAlgs.crossover(father, mother, population{worst2});
        else %mutate

            if using_tools && ~transform 
                % 50 percent chance for tool change.
                if rand <= 0.5
                    % Mutation - tool change.
                    odds = [1 1 0]; % toolchange matchtool both
                    norm_odds = odds ./ sum(odds);
                    cum_odds = cumsum( norm_odds );
                    U = rand;
                    if     U <= cum_odds(1)
                        GAAlgs.mutateToolChange( mother, population{worst1} ); 
                        GAAlgs.mutateToolChange( father, population{worst2} ); 
                    elseif U <= cum_odds(2)
                        GAAlgs.mutateMatchTool( mother, population{worst1} );
                        GAAlgs.mutateMatchTool( father, population{worst2} );
                    elseif U <= cum_odds(3)
                        GAAlgs.mutateToolChange( mother, population{worst1} ); 
                        GAAlgs.mutateToolChange( father, population{worst2} ); 
                        GAAlgs.mutateMatchTool( mother, population{worst1} );
                        GAAlgs.mutateMatchTool( father, population{worst2} );
                    end
                    continue;
                end
            end

            % Regular mutations
            odds = [0 1 2 0]; % internal external reverse transfer
            norm_odds = odds ./ sum(odds);
            cum_odds = cumsum( norm_odds );
            U = rand;
            if     U <= cum_odds(1)       % Mutation - internal swap. Two tasks are swapped within a robots path.
                GAAlgs.mutateInternalSwap( mother, population{worst1} ); 
                GAAlgs.mutateInternalSwap( father, population{worst2} ); 
            elseif U <= cum_odds(2)       % Mutation - external swap. Two tasks are swapped between robots.
                GAAlgs.mutateExternalSwap( mother, population{worst1} ); 
                GAAlgs.mutateExternalSwap( father, population{worst2} );
            elseif U <= cum_odds(3)       % Mutation - reverse. Tasks in a random section are reversed.
                GAAlgs.mutateReverse( mother, population{worst1} ); 
                GAAlgs.mutateReverse( father, population{worst2} ); 
            elseif U <= cum_odds(4)       % Mutation - transfer. A task is given from one robot to another.
                GAAlgs.mutateTransfer( mother, population{worst1} );
                GAAlgs.mutateTransfer( father, population{worst2} );
            end
        end
    end
    

    % Sort the population
    if objective == 1
        costs = cellfun(@(x)x.minisum, population);
    end
    if objective == 2
        costs = cellfun(@(x)x.minimax, population);
    end
    [costs,sort_index] = sort(costs);
    population = population(sort_index);
    
    % Keep track of cost progression.
    cost_tracking(iteration) = costs(1);
        
    % Update the best.
    if costs(1) < best_cost
        best_cost = costs(1);
        best_solution = population{1};
        % Progression display.
        if display_graphs
            figure(1);
            clf;
            population{1}.plotGene(objective);
            pause(0.01);
            % Save the image.
%             fig = gcf;
%             fig.PaperPositionMode = 'auto';
%             fig_pos = fig.PaperPosition;
%             fig.PaperSize = [fig_pos(3) fig_pos(4)];
%             name = ['atb' num2str(fig_counter)];
%             print(fig, '-dpng', '-r150', name);
%             fig_counter = fig_counter + 1;
        end
    end
    
    % Steady state limit break.
%     if prev_cost == costs(1)
%         same_counter = same_counter + 1;
%     else
%         same_counter = 0;
%         prev_cost = costs(1);
%     end
%     if same_counter > 500
%         cost_tracking = cost_tracking(1:iteration);
%         break;
%     end
    
end


%% Select best solution.
% Sort the population
if objective == 1
    costs = cellfun(@(x)x.minisum, population);
end
if objective == 2
    costs = cellfun(@(x)x.minimax, population);
end
[~,sort_index] = sort(costs);
population = population(sort_index);
result = costs(1);

if display_graphs
    figure(1); clf;
    best_solution.plotGene(objective);
end

% Get last iteration that it improved.
[~,ind] = min(cost_tracking);
disp(['Best at iteration: ' num2str(ind)]);

%% Improve using 2-opt
% Improve best.
best_solution = GAAlgs.perform2Opt( best_solution, false, false );
if objective == 1
    result_with_2_opt = best_solution.minisum;
end
if objective == 2
    result_with_2_opt = best_solution.minimax;
end

if display_graphs
    figure(2);
    best_solution.plotGene(objective);
%     fig = gcf;
%     fig.PaperPositionMode = 'auto';
%     fig_pos = fig.PaperPosition;
%     fig.PaperSize = [fig_pos(3) fig_pos(4)];
%     name = ['GAend'];
%     print(fig, '-dpng', '-r150', name);
%     fig_counter = fig_counter + 1;
end

% Show other graphs.
if display_graphs
    figure(3);
    plot(cost_tracking);
    title('Objective Cost');
end
end